"""Generate configuration for Samba AD DC."""

import os
from ipalab_config.utils import copy_resource_files

base_config = {
    "image": "localhost/samba-addc",
    "build": {
        "context": "containerfiles",
        "dockerfile": "external-nodes",
        "args": {"packages": "systemd"},
    },
    "command": "/usr/sbin/init",
}


def gen_config(_lab_config, base_dir, _node, _options):
    """Update node configuration."""
    copy_resource_files(
        "playbooks/deploy_addc.yml", os.path.join(base_dir, "playbooks")
    )



"""Generate configuration for the Unbound nameserver."""

import os
import ipaddress
import textwrap

from ipalab_config.utils import copy_helper_files, copy_extra_files, save_file


base_config = {
    "image": "localhost/unbound",
    "build": {"context": "unbound", "dockerfile": "Containerfile"},
    "volumes": [
        "${PWD}/unbound:/etc/unbound:rw",
    ],
}


def gen_config(lab_config, base_dir, _node, options):
    """Generate configuration for external DNS container."""
    subnet = lab_config["subnet"]
    domains = []
    networks = [ipaddress.IPv4Interface(subnet).network]
    zone_files = []

    zone_template = textwrap.dedent(
        """\
    auth-zone:
        name: {name}
        zonefile: /etc/unbound/zones/{filename}
        for-downstream: yes
        for-upstream: no
    """
    )

    copy_helper_files(base_dir, "unbound")

    zone_data = []
    for zone in options.get("zones", []):
        filename = zone["file"]
        zone_files.append(filename)
        if "name" in zone:
            name = zone["name"]
            domains.append(name)
        else:
            network = ipaddress.IPv4Interface(zone["reverse_ip"]).network
            networks.append(network)
            name = network[0].reverse_pointer.split(".", 1)[1]
        zone_data.append({"name": name, "filename": os.path.basename(filename)})

    save_file(
        base_dir,
        "unbound/zones.conf",
        "\n".join(zone_template.format(**zone) for zone in zone_data),
    )
    save_file(
        base_dir,
        "unbound/domains",
        "\n".join(f"private-domain: {domain}" for domain in domains),
    )
    save_file(
        base_dir,
        "unbound/access_control",
        "\n".join(
            f"access-control: {network} allow" for network in set(networks)
        ),
    )

    copy_extra_files(zone_files, os.path.join(base_dir, "unbound/zones"))



"""ipalab_config utility functions."""

import os
import shutil
import socket
import ipaddress
import importlib
import importlib.resources
import functools

from ipalab_config.logger import logger


def die(msg, err=1):  # pragma: no cover
    """Display message to stderr stream and exit program with error."""
    logger.critical(msg)
    return err


def ensure_fqdn(hostname, domain):
    """Ensure hostame is a FQDN."""
    return hostname if "." in hostname else f"{hostname}.{domain}"


def get_hostname(config, name, domain):
    """Ensure hostname from config is FQDN."""
    hostname = config.get("hostname", name)
    return ensure_fqdn(hostname, domain)


def is_ip_address(addr):
    """Check if a given string represents an IP address."""
    try:
        socket.inet_pton(socket.AF_INET, addr)
    except OSError:
        try:
            socket.inet_pton(socket.AF_INET6, addr)
        except OSError:
            return False
    return True


def get_ip_address_generator(for_cidr=None):
    """Create an IP address generator given a network CIDR."""
    network = ipaddress.IPv4Interface(for_cidr or "192.168.159.0/24").network
    generator = network.hosts()
    next(generator)  # assume first IP is the gateway IP address.
    return generator


def get_service_ip_address(service):
    """Return the compose service IP address."""
    return service["networks"]["ipanet"]["ipv4_address"]


def copy_extra_files(files, target_dir, source=None):
    """Copy files to the target directory."""
    os.makedirs(target_dir, exist_ok=True)
    for arg in files:
        filename = os.path.basename(arg)
        shutil.copyfile(
            os.path.join(source or "", arg),
            os.path.join(target_dir, filename),
        )


def copy_resource_files(files, target_dir):
    """Copy ipalab-config resource files to target directory."""
    os.makedirs(target_dir, exist_ok=True)
    if not isinstance(files, (list, tuple)):
        files = [files]
    for source in files:
        filename = os.path.join(
            importlib.resources.files("ipalab_config"), "data", source
        )
        shutil.copyfile(
            filename, os.path.join(target_dir, os.path.basename(filename))
        )


def copy_helper_files(base_dir, directory, source=None):
    """Copy directory helper files to target directory"""
    target_dir = os.path.join(base_dir, directory)
    os.makedirs(target_dir, exist_ok=True)

    if source is None:
        source = os.path.join(
            importlib.resources.files("ipalab_config"), "data"
        )
    origin = os.path.join(source, directory)
    shutil.copytree(origin, target_dir, dirs_exist_ok=True)


def save_file(base_dir, filename, data):
    """Write data to an output file."""
    # pylint: disable=unspecified-encoding
    with open(os.path.join(base_dir, filename), "w") as out:
        out.write(data)


@functools.cache
def import_external_role_module(role):
    """Import an external role module."""
    role_mod = f"ipalab_config.external_role.{role}"
    return importlib.import_module(role_mod)



"""Generate configuration for external host: Keycloak"""

import textwrap

from ipalab_config.utils import copy_helper_files, save_file

base_config = {
    "image": "localhost/keycloak",
    "build": {
        "context": "keycloak",
        "dockerfile": "Containerfile",
        "args": {"hostname": "{hostname}"},
    },
    "environment": {
        "KC_BOOTSTRAP_ADMIN_USERNAME": "{admin_username}",
        "KC_BOOTSTRAP_ADMIN_PASSWORD": "{admin_password}",
        "KC_HOSTNAME": "{hostname}",
    },
    "entrypoint": "/opt/keycloak/bin/kc.sh start",
}


def gen_config(_lab_config, base_dir, node, options):
    """Update node configuration."""

    def apply_formatting(config_section, keys, defaults):
        for key in keys:
            config_section[key] = config_section[key].format(**defaults)
        return config_section

    defaults = {
        "admin_username": "admin",
        "admin_password": "secret123",
        "oidc_password": "secret123",
        "hostname": node.get("hostname", "localhost"),
    } | (options or {})

    # Then in gen_config:
    env_keys = [
        "KC_BOOTSTRAP_ADMIN_USERNAME",
        "KC_BOOTSTRAP_ADMIN_PASSWORD",
        "KC_HOSTNAME",
    ]

    node["environment"] = apply_formatting(
        node["environment"], env_keys, defaults
    )

    node["build"]["args"] = apply_formatting(
        node["build"]["args"], ["hostname"], defaults
    )

    copy_helper_files(base_dir, "keycloak")

    keycloak_config = textwrap.dedent(
        f"""\
    ADMIN="{defaults['admin_username']}"
    PASSWORD="{defaults['admin_password']}"
    OIDCPASSWORD="{defaults['oidc_password']}"
    KEYCLOAK="{node.get('hostname', 'localhost1')}"
    KEYCLOAK_CONTAINER="{node['container_name']}"

    KEYCLOAK_URL="https://${{KEYCLOAK}}:8443"
    TRUSTPASSWORD="password"
    """
    )

    save_file(base_dir, "keycloak/keycloak_config.sh", keycloak_config)



"""Helper functions to generate an Ansible YAML inventory file."""

from ipalab_config.utils import die, get_hostname, ensure_fqdn


def get_node_name(name, deployment):
    """Return the proper name to use for the node."""
    if deployment["container_fqdn"]:
        return ensure_fqdn(name, deployment["domain"])
    return name


def get_server_inventory(config, default_config, deployment):
    """Get inventory configuration for the ipaserver"""
    cap_opts = {
        "DNS": {
            "ipaserver_setup_dns": True,
            "ipaserver_auto_forwarders": True,
            "ipaserver_forward_policy": "first",
            "ipaserver_no_dnssec_validation": True,
            "ipaserver_auto_reverse": True,
        },
        "KRA": {
            "ipaserver_setup_kra": True,
        },
        "AD": {"ipaserver_setup_adtrust": True},
        "RSN": {"ipaserver_random_serial_numbers": True},
    }

    name = get_node_name(config["name"], deployment)
    hostname = get_hostname(config, name, deployment["domain"])
    options = (
        {}
        if config.get("no_limit_uid")
        else {
            "ipaserver_idstart": 60001,
            "ipaserver_idmax": 62000,
            "ipaserver_rid_base": 63000,
            "ipaserver_secondary_rid_base": 65000,
        }
    )
    options.update({"ipaserver_hostname": hostname, **default_config})
    for cap in config.get("capabilities", []):
        options.update(cap_opts.get(cap, {}))
    options.update(
        {
            "ipaclient_no_ntp": False,
            "ipaserver_setup_firewalld": False,
            "ipaserver_no_host_dns": True,
        }
    )
    options.update(dict(config.get("vars", {})))
    return {name: options}


def get_replicas_inventory(replicas_config, default_config, deployment):
    """Get inventory configuration for the ipaserver"""
    if not replicas_config:
        return {}
    replicas = {}
    cap_opts = {
        "DNS": {
            "ipareplica_setup_dns": True,
            "ipareplica_auto_forwarders": True,
            "ipareplica_forward_policy": "first",
            "ipareplica_no_dnssec_validation": True,
            "ipareplica_auto_reverse": True,
        },
        "KRA": {
            "ipareplica_setup_dns": True,
        },
        "AD": {"ipareplica_setup_adtrust": True},
        "CA": {"ipareplica_setup_ca": True},
        "HIDDEN": {"ipareplica_hidden_replica": True},
    }

    # common options
    common = {
        "ipareplica_setup_firewalld": False,
        "ipareplica_no_host_dns": True,
        **default_config,
    }
    server = deployment.get("server")
    if server:
        common["ipareplica_servers"] = server["ipaserver_hostname"]
    domain = deployment["domain"]
    for replica in replicas_config:
        name = get_node_name(replica["name"], deployment)
        hostname = get_hostname(replica, name, domain)
        options = {"ipareplica_hostname": hostname, **common}
        for cap in replica.get("capabilities", []):
            options.update(cap_opts.get(cap, {}))
        options.update(dict(replica.get("vars", {})))
        replicas[name] = options
    return replicas


def get_clients_inventory(config, default_config, deployment):
    """Get inventory configuration for the ipaserver"""
    clients = {}
    common = {**default_config}
    server = deployment.get("server")
    if server:
        common["ipaadmin_password"] = server["ipaadmin_password"]
        common["ipaclient_servers"] = server["ipaserver_hostname"]
    if isinstance(config, dict):
        client_list = config.get("hosts", [])
        common.update(config.get("vars", {}))
    else:
        client_list = config
    if not client_list:
        return {}
    for client in client_list or []:
        name = get_node_name(client["name"], deployment)
        hostname = get_hostname(client, name, deployment["domain"])
        clients[name] = {"ipaclient_hostname": hostname, **common}
        clients[name].update(dict(client.get("vars", {})))
    return clients


def gen_inventory_external_nodes(lab_config, lab):
    """Create inventory configuration for IPA external nodes."""
    external = lab_config.get("external", {})
    if not external.get("hosts"):
        return
    hosts = lab.setdefault("hosts", {})
    external_inv = (
        lab.setdefault("children", {})
        .setdefault("external", {})
        .setdefault("hosts", {})
    )
    if "vars" in external:
        external_inv["vars"] = external["vars"]
    for node in external["hosts"]:
        group = lab.setdefault("children", {}).setdefault(
            f"role_{node.get("role", "none")}", {"hosts": {}}
        )
        variables = node.get("vars", {})
        hosts[node["name"]] = dict(variables) if variables else None
        group["hosts"].update({node["name"]: None})
        external_inv.update({node["name"]: None})


def gen_inventory_ipa_deployments(lab_config, lab):
    """Create inventory configuration for IPA deployments."""

    def add_hosts_to_inventory(lab_name, lab, item, data):
        # add to host list
        hosts = lab.setdefault("hosts", {})
        hosts.update(data)
        # groups
        lab_child = lab.setdefault("children", {})
        # add to group list
        group = lab_child.setdefault(item, {}).setdefault("hosts", {})
        group.update({item: None for item in data})
        # fmt: off
        # add to deployment list
        deployment_inv = (
            lab_child.setdefault(lab_name, {}).setdefault("hosts", {})
        )
        # fmt: on
        deployment_inv.update({item: None for item in data})

    ipa_deployments = lab_config.setdefault("ipa_deployments", [])
    deployment_dns = lab_config["deployment_nameservers"]
    if len(ipa_deployments) > 1 and not lab_config["container_fqdn"]:
        raise ValueError(
            "With multiple IPA deployments, "
            "'container_fqdn' must be set to 'true'."
        )
    for deployment, nameservers in zip(ipa_deployments, deployment_dns):
        name = deployment["name"]
        # process deployment
        domain = deployment.setdefault("domain", lab_config.get("domain"))
        deployment.setdefault("subnet", lab_config["subnet"])
        deployment.setdefault("container_fqdn", lab_config["container_fqdn"])
        default_config = {
            "ipaadmin_password": deployment.get(
                "admin_password", "SomeADMINpassword"
            ),
            "ipadm_password": deployment.get("dm_password", "SomeDMpassword"),
            "ipaserver_realm": deployment.get("realm", domain).upper(),
            "ipaclient_no_ntp": True,
            **deployment.get("vars", {}),
        }
        if domain and "ipaserver_domain" not in default_config:
            default_config["ipaserver_domain"] = domain
        cluster_config = deployment.get("cluster")
        if not cluster_config:  # pragma: no cover
            die(f"Cluster not defined for domain '{domain}'")
        # parse first server
        servers = cluster_config.get("servers")
        if servers:
            server = get_server_inventory(
                servers[0], default_config, deployment
            )
            add_hosts_to_inventory(name, lab, "ipaserver", server)
            # set deployment first server
            deployment["server"] = next(iter(server.values()))
            if deployment["server"].get("ipaserver_setup_dns", False):
                default_config.update(
                    {
                        "ipaclient_cleanup_dns_resolver": True,
                        "ipaclient_configure_dns_resolver": True,
                        "ipaclient_dns_servers": nameservers,
                    }
                )
            # parse replicas
            replicas = get_replicas_inventory(
                servers[1:],
                default_config,
                deployment,
            )
            if replicas:
                add_hosts_to_inventory(name, lab, "ipareplicas", replicas)
        # parse clients
        clients_config = cluster_config.get("clients")
        client_base_config = {
            key: value
            for key, value in default_config.items()
            if not any(map(key.startswith, ["ipaserver_", "ipadm_"]))
        }
        client_base_config.update({"ipaserver_domain": domain})
        clients = get_clients_inventory(
            clients_config, client_base_config, deployment
        )
        if clients:
            add_hosts_to_inventory(name, lab, "ipaclients", clients)
        # add deployment to inventory
        lab["children"].setdefault("ipa_deployments", {"children": {}})
        lab["children"]["ipa_deployments"]["children"][name] = None
    return ipa_deployments


def gen_inventory_data(lab_config):
    """Generate inventory file based on provided configuration"""
    labname = lab_config["lab_name"]
    lab = {"vars": {"ansible_connection": "podman"}}
    gen_inventory_external_nodes(lab_config, lab)
    gen_inventory_ipa_deployments(lab_config, lab)
    node_names = list(lab_config["ipa_deployments"])
    node_names.extend(list(lab.get("external", {}).keys()))
    if any(
        labname == deployment["name"]
        for deployment in lab_config["ipa_deployments"]
    ):
        raise ValueError(f"Deployment name must be unique: {labname}")

    return {labname.replace("-", "_"): lab}



"""Generate compose and inventory configuration for a FreeIPA cluster."""

import argparse
import os
import sys

from ruamel.yaml import YAML

try:
    from jinja2 import Environment

    HAS_JINJA = True
except ImportError:
    HAS_JINJA = False

from ipalab_config import __version__
from ipalab_config.utils import (
    die,
    copy_extra_files,
    copy_helper_files,
    save_file,
    get_service_ip_address,
    import_external_role_module,
)
from ipalab_config.compose import gen_compose_data
from ipalab_config.inventory import gen_inventory_data


def parse_arguments():
    """Parse command line arguments."""
    opt_parser = argparse.ArgumentParser(
        prog="ipa-lab-config",
        description=(
            "Generate compose and inventory configuration for FreeIPA lab."
        ),
    )
    opt_parser.add_argument("CONFIG", help="Lab description")
    opt_parser.add_argument("--version", action="version", version=__version__)
    opt_parser.add_argument(
        "-o",
        "--output",
        dest="OUTPUT",
        metavar="OUTPUT",
        default=None,
        help="Output directory",
    )
    opt_parser.add_argument(
        "-f",
        "--file",
        dest="RECIPES",
        metavar="CONTAINERFILE",
        action="append",
        default=[],
        help=(
            "Containerfile to use for creating images. "
            "May be used more than once for multiple files."
        ),
    )
    opt_parser.add_argument(
        "-p",
        "--playbook",
        dest="PLAYBOOKS",
        metavar="PLAYBOOKS",
        action="append",
        default=[],
        help=(
            "Add playbook to the resulting configuration 'playbooks' "
            "directory. When adding a directory, all 'yml' or 'yaml' "
            "files in the directory, searched recursively will be copied. "
            "May be used more than once for multiple files."
        ),
    )
    opt_parser.add_argument(
        "-d",
        "--distro",
        dest="DISTRO",
        metavar="DISTRO",
        default=None,
        nargs="?",
        help=(
            "Override default '<distro>:<tag>'. Available distros: "
            "'fedora', 'centos', 'ubuntu'. Any tag supported "
            "by the original distro can be used (default tag is 'latest')."
        ),
    )
    opt_parser.add_argument(
        "--mount-varlog",
        dest="VARLOG",
        action="store_true",
        help=(
            "Associate all nodes '/var/log' to local directories so log "
            "data is accessible through the host. Use 'nolog: true' on "
            "a node configuration to disable the behavior for that node."
        ),
    )
    opt_parser.add_argument(
        "--debug",
        action="store_true",
        help="Run ipalab-config in debug mode.",
    )
    return opt_parser.parse_args()


def save_data(yaml, base_dir, filename, yamldata):
    """Save YAML data as a YAML file."""
    # pylint: disable=unspecified-encoding
    with open(os.path.join(base_dir, filename), "w") as out:
        yaml.dump(yamldata, out)


def gen_external_node_configuration(lab_config, base_dir, compose_config):
    """Generate configuration for external nodes"""
    for _, node_data in compose_config["services"].items():
        external_data = node_data.pop("external_node", None)
        if external_data:
            # update dns on nodes
            if "dns" in node_data:
                dns = node_data["dns"]
                if dns in lab_config["nodes"]:
                    dns = lab_config["nodes"][dns]
                elif dns in compose_config["services"]:
                    dns = get_service_ip_address(
                        compose_config["services"][dns]
                    )
                node_data["dns"] = dns
                if not node_data.get("dns_search"):
                    node_data.pop("dns_search", None)
            else:
                node_data.pop("dns_search", None)
            # update roles
            role = external_data.get("role")
            if role:
                options = external_data.get("options", {})
                try:
                    module = import_external_role_module(role)
                    config_fn = getattr(module, "gen_config", None)
                except ImportError:
                    config_fn = None
                if config_fn:
                    config_fn(lab_config, base_dir, node_data, options)
                else:
                    raise ValueError(
                        f"Role does not provide 'gen_config': '{role}'"
                    )


def gen_optional_files(lab_config, base_dir, yaml):
    """Save optional 'misc' files."""
    # save /etc/hosts file patch
    save_file(
        base_dir,
        "hosts",
        f"\n# ipalab-config hosts for '{lab_config['lab_name']}'\n"
        + "\n".join(
            [
                f"{v:18s}{k.replace('_', '.')}"
                for k, v in lab_config.get("nodes", {}).items()
            ]
        ),
    )

    # add Ansible Galaxy requirements.yml
    save_data(
        yaml,
        base_dir,
        "requirements.yml",
        {
            "collections": [
                {"name": "containers.podman"},
                {"name": "freeipa.ansible_freeipa"},
            ]
        },
    )


def save_containers_data(lab_config, base_dir, args):
    """Copy containerfiles to result directory."""
    copy_helper_files(base_dir, "containerfiles")

    containerfiles = [
        (
            os.path.realpath(
                os.path.join(
                    os.path.dirname(os.path.realpath(args.CONFIG)),
                    containerfile,
                )
            )
            if not containerfile.startswith("/")
            else containerfile
        )
        for containerfile in lab_config.get("containerfiles", [])
    ]
    copy_extra_files(
        containerfiles + args.RECIPES,
        os.path.join(base_dir, "containerfiles"),
    )


def save_ansible_data(_lab_config, base_dir, args):
    """Copy Ansible playbooks to result directory."""
    plays = []
    for play in args.PLAYBOOKS:
        if os.path.isfile(play):
            plays.append(play)
        if os.path.isdir(play):
            plays.extend(
                [
                    os.path.join(dirname, name)
                    for dirname, _, filenames in os.walk(play)
                    for name in filenames
                    if name.endswith(".yml") or name.endswith(".yaml")
                ]
            )
    if plays:
        copy_extra_files(plays, os.path.join(base_dir, "playbooks"))


def set_default_values(data, args):
    """Ensure sane configuration default values."""
    data.setdefault("lab_name", "ipa-lab")
    data.setdefault("distro", "fedora")
    if args.DISTRO:
        distro, *tag = args.DISTRO.split(":", 1)
        data["distro"] = distro
        data["tag"] = "".join(tag) if tag else None
    data.setdefault("container_fqdn", False)
    data.setdefault("mount_varlog", args.VARLOG)
    data.setdefault("domain", "ipalab.local")


def generate_ipalab_configuration():
    """Generate compose and inventory."""
    args = parse_arguments()
    yaml = YAML()
    yaml.explicit_start = True
    yaml.indent(mapping=2, sequence=4, offset=2)

    # pylint: disable=unspecified-encoding
    if not (os.path.isfile(args.CONFIG) and os.access(args.CONFIG, os.R_OK)):
        raise RuntimeError(f"Cannot read config file: {args.CONFIG}")

    with open(args.CONFIG, "r") as config_file:
        source = config_file.read()
        if HAS_JINJA:
            source = Environment().from_string(source).render(ENV=os.environ)
        data = yaml.load(source)

    set_default_values(data, args)

    base_dir = args.OUTPUT or data["lab_name"]

    # generate configuration
    compose_config = gen_compose_data(data)
    inventory_config = gen_inventory_data(data)

    # save configuration
    os.makedirs(base_dir, exist_ok=True)

    gen_external_node_configuration(data, base_dir, compose_config)

    save_data(yaml, base_dir, "compose.yml", compose_config)
    # create log directories
    if data.get("mount_varlog"):
        for node in compose_config["services"]:
            os.makedirs(os.path.join(base_dir, "logs", node), exist_ok=True)

    save_data(yaml, base_dir, "inventory.yml", inventory_config)

    save_containers_data(data, base_dir, args)
    save_ansible_data(data, base_dir, args)
    gen_optional_files(data, base_dir, yaml)

    # process user extra_data
    if "extra_data" in data:
        cwd = os.path.dirname(args.CONFIG)
        for helper in data["extra_data"]:
            if os.path.isabs(helper):
                source = os.path.dirname(helper)
                helper = os.path.basename(helper)
            else:
                source = cwd
            if os.path.isfile(os.path.join(source, helper)):
                copy_extra_files(
                    [helper],
                    os.path.join(base_dir, os.path.dirname(helper)),
                    source=source,
                )
            else:
                copy_helper_files(base_dir, helper, source=cwd)


def main():
    """Trap execution exceptions."""
    debug = "--debug" in sys.argv
    try:
        generate_ipalab_configuration()
    except (  # pragma: no cover
        ValueError,
        FileNotFoundError,
        RuntimeError,
    ) as err:
        if debug:
            raise err from None
        return die(str(err))
    return 0


if __name__ == "__main__":  # pragma: no cover
    main()



"""Provide logging facilities."""

import logging


logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.DEBUG)
logging.addLevelName(logging.CRITICAL, "FATAL")
logger = logging.getLogger("ipalab-config")



"""Helper functions to generate a podman compose file."""

from collections import namedtuple

from ruamel.yaml.scalarstring import DoubleQuotedScalarString
from ruamel.yaml.comments import CommentedMap

from ipalab_config.logger import logger
from ipalab_config.utils import (
    die,
    get_hostname,
    is_ip_address,
    ensure_fqdn,
    get_ip_address_generator,
    import_external_role_module,
)


IP_GENERATOR = get_ip_address_generator()

# Use namedtuple as a class
Network = namedtuple("Network", ["domain", "name", "dns"])


def get_effective_nameserver(nameserver, domain):
    """Return the effective nameserver."""
    if nameserver and not (is_ip_address(nameserver) or "{" in nameserver):
        return f"{{{ensure_fqdn(nameserver, domain)}}}"
    return nameserver


def get_node_base_config(  # pylint: disable=R0913,R0917
    name, hostname, networkname, ipaddr, distro=None, tag=None, image=None
):
    """Returns the basic node configuration."""
    node_image = {}
    if distro is None and image is None:
        distro = "fedora"
    if distro:
        node_image = {
            "image": f"localhost/{distro}:{'latest' if tag is None else tag}",
            "build": {
                "context": "containerfiles",
                "dockerfile": f"{distro}",
            },
        }
    if image:
        node_image = {"image": image}

    # fmt: off
    result = CommentedMap({
        "container_name": name,
        # Use DoubleQuotedScalarString as 'no' without quotes may be
        # interpreted as boolean False by PyYAML loaders.
        "restart": DoubleQuotedScalarString("no"),
        "cap_add": ["SYS_ADMIN", "DAC_READ_SEARCH"],
        "security_opt": ["label=disable"],
        "hostname": hostname,
        "networks": {networkname: {"ipv4_address": str(ipaddr)}},
        **node_image
    })
    # fmt: on
    supported_distros = {"fedora", "centos", "external-nodes", "ubuntu"}
    if "build" in result:
        if tag is not None and distro in supported_distros:
            args = {"distro_image": distro, "distro_tag": tag}
            result["build"]["args"] = args
        else:
            # If no tag is given, and distro is one of the ipalab-config
            # provided ones, add a commented our "args" option to "build".
            if distro in supported_distros:
                result.yaml_set_comment_before_after_key(
                    "build",
                    after=(
                        "You may set the desired distro/version setting:\n"
                        f"args: {{distro_image: {distro}, distro_tag: latest}}"
                    ),
                    after_indent=6,
                )
    return result


def get_container_name(container, domain, container_fqdn):
    """Ensure proper creation of container name."""
    name = container["name"]
    if container_fqdn:
        if container.get("hostname"):
            name = ensure_fqdn(container["hostname"], domain)
        else:
            name = ensure_fqdn(name, domain)
    return name


def get_compose_config(containers, ips=IP_GENERATOR, **kwargs):
    """Create config for all containers in the list."""

    def node_dns_key(hostname):
        return hostname.replace(".", "_")

    if isinstance(containers, dict):
        containers = containers.get("hosts", [])
    if not containers:
        return {}, {}
    result = {}
    nodes = {}
    network = kwargs.get("network")
    if network is None:
        raise RuntimeError("Node network not defined.")
    container_fqdn = kwargs.get("container_fqdn", False)
    distro = kwargs.get("distro", "fedora")
    tag = kwargs.get("tag")
    mount_varlog = kwargs.get("mount_varlog", False)
    for container in containers:
        name = get_container_name(container, network.domain, container_fqdn)
        node_distro = container.get("distro", distro)
        node_tag = container.get("tag", tag)
        node_image = container.get("image")
        hostname = get_hostname(container, name, network.domain)
        ipaddr = container.get("ip_address")
        if not ipaddr:
            ipaddr = next(ips)
        nodes[node_dns_key(hostname)] = str(ipaddr)
        config = get_node_base_config(
            name,
            hostname,
            network.name,
            str(ipaddr),
            node_distro,
            node_tag,
            node_image,
        )
        if "memory" in container:
            config.update(
                {"mem_limit": container["memory"].lower(), "memory_swap": -1}
            )
        # DNS
        effective_dns = get_effective_nameserver(
            container.get("dns", network.dns), network.domain
        )
        if effective_dns:
            config["dns"] = (
                effective_dns
                if is_ip_address(effective_dns)
                else node_dns_key(effective_dns)
            )
        config["dns_search"] = network.domain

        volumes = container.get("volumes", [])
        if not isinstance(volumes, (list, tuple)):
            volumes = [volumes]
        if mount_varlog and not container.get("nolog", False):
            volumes.extend([f"${{PWD}}/logs/{name}:/var/log:rw"])
        if volumes:
            config.update({"volumes": volumes})

        result[name] = config
    return nodes, result


def get_network_config(lab_config):
    """Generate the compose "networks" configuration."""
    subnet = lab_config.get("subnet")
    network = lab_config.get("network") or {}
    if isinstance(network, str):
        networkname = network
        config = {networkname: {"external": True}}
    else:
        networkname = network.get("name", "ipanet")
        name = network.get("name", f"ipanet-{lab_config['lab_name']}")
        subnet = network.get("subnet", subnet or "192.168.159.0/24")
        config = {
            networkname: {
                "name": name,
                "driver": network.get("driver", "bridge"),
                "ipam": {"config": [{"subnet": subnet}]},
            }
        }
        no_dns = network.get("no_dns", False)
        if network.get("external"):
            config[networkname]["external"] = True
        if no_dns or "dns" in network:
            logger.warning("DNS options require 'podman-compose > 1.3.0'")
        if no_dns:
            config[networkname]["x-podman.disable_dns"] = True
        if "dns" in network:
            dns = network["dns"]
            if not dns:
                raise ValueError("'network.dns' must not be empty")
            if not isinstance(dns, (str, list)):
                raise ValueError("'network.dns' must be a string or a list")
            config[networkname]["x-podman.dns"] = dns
    # Ensure subnet is correctly set
    if config[networkname].get("external", False) and subnet is None:
        raise ValueError("'subnet' is required for 'external' networks")
    lab_config.setdefault("subnet", subnet)
    return networkname, config


def get_ipa_deployments_configuration(lab_config, networkname, ip_generator):
    """Generate compose configuration for all IPA deployments."""
    lab_config.setdefault("deployment_nameservers", [])
    labdns = lab_config.get("dns")
    labdomain = lab_config.get("domain")
    services = {}
    for deployment in lab_config.setdefault("ipa_deployments", []):
        domain = deployment.get("domain", labdomain)
        dns = deployment.get("dns", labdns)
        if dns and not is_ip_address(dns):
            # pylint: disable=consider-using-f-string
            dns = "{{{0}}}".format(ensure_fqdn(dns, domain))
        config = {
            "network": Network(
                domain,
                networkname,
                get_effective_nameserver(dns, domain),
            ),
            "container_fqdn": lab_config["container_fqdn"],
            "distro": deployment.get("distro", lab_config["distro"]),
            "tag": deployment.get("tag", lab_config.get("tag")),
            "mount_varlog": lab_config.get("mount_varlog", False),
        }
        cluster_config = deployment.get("cluster")
        if not cluster_config:
            die(f"Cluster not defined for domain '{domain}'")
        nodes = {}
        # Get servers configurations
        servers = cluster_config.get("servers")
        if servers:
            ips, servers_cfg = get_compose_config(
                servers, ip_generator, **config
            )
            deployment_dns = [
                "{{{0}}}".format(  # pylint: disable=consider-using-f-string
                    ensure_fqdn(host["name"], domain)
                )
                .replace(".", "_")
                .format(**ips)
                for host in servers
                if "DNS" in host.get("capabilities", [])
            ]
            lab_config["deployment_nameservers"].append(deployment_dns)
            services.update(servers_cfg)
            nodes.update(ips)
        else:
            logger.warning(
                "No servers defined for domain '%(domain)s'",
                domain=domain,
            )
            lab_config["deployment_nameservers"].append(None)
        # Get clients configuration
        clients = cluster_config.get("clients")
        ips, clients_cfg = get_compose_config(clients, ip_generator, **config)
        services.update(clients_cfg)
        nodes.update(ips)
        # We must have at lest one node at the end.
        if not nodes:
            die("At least one server or client must be defined for {domain}.")
        # update nodes list
        lab_config.setdefault("nodes", {}).update(nodes)
        # Update 'dns' on each service
        for service in services.values():
            if "dns" in service:
                service["dns"] = service["dns"].format(**nodes)
            else:
                service.pop("dns_search", None)

    return services


def get_external_hosts_configuration(lab_config, networkname, ip_generator):
    """Generate configuration for hosts external to IPA deployments."""
    external = lab_config.get("external", {})
    services = {}
    node_config = {
        "network": Network(
            external.get("domain", lab_config.get("domain", "ipalab.local")),
            networkname,
            get_effective_nameserver(lab_config.get("dns", ""), ""),
        ),
        "container_fqdn": lab_config["container_fqdn"],
        "distro": "external-nodes",
        "mount_varlog": lab_config["mount_varlog"],
    }
    ext_nodes = external.get("hosts", [])
    nodes, services = get_compose_config(ext_nodes, ip_generator, **node_config)
    # update nodes list
    lab_config.setdefault("nodes", {}).update(nodes)

    # Handle DNS specially
    dns = None
    for node in ext_nodes:
        if node.get("role") == "dns":
            service = services[node["name"]]
            ip_address = service["networks"][networkname]["ipv4_address"]
            # Use ugly side effect
            dns = lab_config["dns"] = ip_address
            # Use first nameserver as lab nameserver
            break

    # Udate external nodes
    for node in ext_nodes:
        role = node.get("role")

        def service_from_node(services, node):
            if node["name"] in services:
                return services[node["name"]]
            if node.get("hostname") in services:
                return services[node.get("hostname")]
            return None

        service = service_from_node(services, node)
        if role:
            # Use defaults for choosen role
            try:
                module = import_external_role_module(role)
            except ImportError:
                raise RuntimeError(f"Invalid external role: {role}") from None
            service.update(getattr(module, "base_config", {}))
        # Use provided node image
        if node.get("image"):
            service.pop("distro", None)
            service["image"] = node.get("image")
        # Merge volumes with user configuration
        volumes = service.pop("volumes", None)
        if volumes:
            uservol = service.get("volumes", [])
            uservol.extend(volumes)
            service["volumes"] = uservol
        # save node original configuration
        service["external_node"] = node
        # setup node DNS
        if dns:
            service["dns"] = dns

    return services


def gen_compose_data(lab_config):
    """Generate podamn compose file based on provided configuration."""
    config = {"name": lab_config["lab_name"]}
    config.setdefault("services", {})

    networkname, config["networks"] = get_network_config(lab_config)
    ip_generator = get_ip_address_generator(lab_config["subnet"])

    # Add external hosts
    config["services"].update(
        get_external_hosts_configuration(lab_config, networkname, ip_generator)
    )
    # Add IPA deployments
    config["services"].update(
        get_ipa_deployments_configuration(lab_config, networkname, ip_generator)
    )

    return config
